@@ -2,63 +2,10 @@
 #![doc = include_str!("../README.md")]
 
 use std::{
-    cmp::max, cmp::Reverse, default::Default, fmt, iter::FromIterator, iter::FusedIterator,
-    mem::swap, num::Wrapping,
+    cmp::Reverse, default::Default, fmt, iter::FromIterator, iter::FusedIterator, num::Wrapping,
 };
 
-#[derive(Clone)]
-struct Bucket<K, V> {
-    max: Option<K>,
-    elems: Vec<(K, V)>,
-}
-
-impl<K, V> Bucket<K, V> {
-    fn is_empty(&self) -> bool {
-        self.elems.is_empty()
-    }
-
-    fn drain(&mut self) -> std::vec::Drain<(K, V)> {
-        self.max = None;
-        self.elems.drain(..)
-    }
-
-    fn iter(&self) -> std::slice::Iter<(K, V)> {
-        self.elems.iter()
-    }
-
-    fn into_iter(self) -> std::vec::IntoIter<(K, V)> {
-        self.elems.into_iter()
-    }
-
-    fn clear(&mut self) {
-        self.max = None;
-        self.elems.clear();
-    }
-
-    fn shrink_to_fit(&mut self) {
-        self.elems.shrink_to_fit();
-    }
-}
-
-impl<K: Ord + Copy, V> Bucket<K, V> {
-    fn push(&mut self, key: K, value: V) {
-        self.max = max(self.max, Some(key));
-        self.elems.push((key, value));
-    }
-
-    fn pop(&mut self) -> Option<(K, V)> {
-        self.elems.pop()
-    }
-}
-
-impl<K, V> Default for Bucket<K, V> {
-    fn default() -> Bucket<K, V> {
-        Bucket {
-            max: None,
-            elems: Vec::new(),
-        }
-    }
-}
+type Bucket<K, V> = Vec<(K, V)>;
 
 /// A montone priority queue implemented using a radix heap.
 ///
@@ -133,32 +80,42 @@ impl<K: Radix + Ord + Copy, V> RadixHeapMap<K, V> {
         self.top = Some(top);
     }
 
-    #[inline]
-    fn push_nocheck(&mut self, key: K, value: V, top: K) {
-        self.buckets[key.radix_distance(&top) as usize].push(key, value);
-    }
+    /// Sets the top value to the current maximum key value in the heap
+    pub fn constrain(&mut self) {
+        let repush = if self.top.is_some() {
+            let index = self
+                .buckets
+                .iter()
+                .enumerate()
+                .find(|&(_, bucket)| !bucket.is_empty())
+                .map(|(i, _)| i);
 
-    #[inline]
-    fn repush_bucket<F>(&mut self, mut bucket: F)
-    where
-        F: FnMut(&mut Self) -> &mut Bucket<K, V>,
-    {
-        let mut repush = Bucket::default();
+            match index {
+                None | Some(0) => None,
+                Some(index) => {
+                    let (buckets, rest) = self.buckets.split_at_mut(index);
+                    Some((buckets, &mut rest[0]))
+                }
+            }
+        } else if !self.initial.is_empty() {
+            Some((&mut self.buckets[..], &mut self.initial))
+        } else {
+            None
+        };
 
-        swap(bucket(self), &mut repush);
-        let top = repush.max.expect("Expected non-empty bucket");
-        self.top = Some(top);
+        if let Some((buckets, repush)) = repush {
+            let top = *repush
+                .iter()
+                .map(|(k, _)| k)
+                .max()
+                .expect("Expected non-empty bucket");
 
-        for (k, v) in repush.drain() {
-            self.push_nocheck(k, v, top);
+            self.top = Some(top);
+
+            repush.drain(..).for_each(|(key, value)| {
+                buckets[key.radix_distance(&top) as usize].push((key, value))
+            });
         }
-
-        // Swap `repush` back again (purely to save memory allocation,
-        // they are both empty at this point but `repush` has some
-        // memory allocated)
-        swap(bucket(self), &mut repush);
-
-        debug_assert!(repush.is_empty());
     }
 
     /// Pushes a new key value pair onto the heap.
@@ -168,36 +125,17 @@ impl<K: Radix + Ord + Copy, V> RadixHeapMap<K, V> {
     /// Panics if the key is larger than the current top key.
     #[inline]
     pub fn push(&mut self, key: K, value: V) {
-        if let Some(top) = self.top {
+        let bucket = if let Some(top) = self.top {
             assert!(key <= top, "Key must be lower or equal to current top key");
-            self.push_nocheck(key, value, top);
+            &mut self.buckets[key.radix_distance(&top) as usize]
         } else {
-            self.initial.push(key, value);
-        }
+            &mut self.initial
+        };
 
+        bucket.push((key, value));
         self.len += 1;
     }
 
-    /// Sets the top value to the current maximum key value in the heap
-    pub fn constrain(&mut self) {
-        if self.top.is_some() {
-            let index = self
-                .buckets
-                .iter()
-                .enumerate()
-                .find(|&(_, bucket)| !bucket.is_empty())
-                .map(|(i, _)| i);
-
-            if let Some(index) = index {
-                if index != 0 {
-                    self.repush_bucket(|x| &mut x.buckets[index]);
-                }
-            }
-        } else if !self.initial.is_empty() {
-            self.repush_bucket(|x| &mut x.initial);
-        }
-    }
-
     /// Remove the greatest element from the heap and returns it, or `None` if
     /// empty.
     ///
@@ -207,33 +145,37 @@ impl<K: Radix + Ord + Copy, V> RadixHeapMap<K, V> {
     /// This will set the top key to the extracted key.
     #[inline]
     pub fn pop(&mut self) -> Option<(K, V)> {
-        self.constrain();
+        let mut constrained = false;
 
-        let pop = self
-            .buckets
-            .first_mut()
-            .expect("Expected at least one bucket")
-            .pop();
+        loop {
+            let pop = self.buckets[0].pop();
 
-        if pop.is_some() {
-            self.len -= 1;
+            if pop.is_some() {
+                self.len -= 1;
+                return pop;
+            } else if constrained {
+                return pop;
+            } else {
+                constrained = true;
+                self.constrain()
+            }
         }
-
-        pop
     }
 
     /// Returns the number of elements in the heap
+    #[inline]
     pub fn len(&self) -> usize {
         self.len
     }
 
     /// Returns true if there is no elements in the heap
+    #[inline]
     pub fn is_empty(&self) -> bool {
         self.len() == 0
     }
 
-    /// The current top value. All keys pushed onto the heap must be smaller
-    /// than this value.
+    /// The current top value. All keys pushed onto the heap must be smaller than this value.
+    #[inline]
     pub fn top(&self) -> Option<K> {
         self.top
     }
@@ -521,12 +463,14 @@ radix_int_impl!(i8);
 radix_int_impl!(i16);
 radix_int_impl!(i32);
 radix_int_impl!(i64);
+radix_int_impl!(i128);
 radix_int_impl!(isize);
 
 radix_int_impl!(u8);
 radix_int_impl!(u16);
 radix_int_impl!(u32);
 radix_int_impl!(u64);
+radix_int_impl!(u128);
 radix_int_impl!(usize);
 
 #[cfg(feature = "ordered-float")]
@@ -789,6 +733,8 @@ mod tests {
         quickcheck(prop as fn(Vec<u8>) -> bool);
         quickcheck(prop as fn(Vec<i16>) -> bool);
         quickcheck(prop as fn(Vec<(i64, usize)>) -> bool);
+        quickcheck(prop as fn(Vec<i128>) -> bool);
+        quickcheck(prop as fn(Vec<u128>) -> bool);
     }
 
     #[cfg(feature = "ordered-float")]
